% Chemie.pl - Prekladac chemickeho nazvoslovi, Copyright(C) 2003 Jan Curn.
% Program odladen v SWI-Prologu (www.swi-prolog.org)
%
% Hlavni predikat: 
%     dejnazev(+ChemVyraz, -Nazev)
%
% Priklady: 
%     dejnazev('H2SO4', X).       X='kyselina sirova'
%     dejnazev('CaMg(CO3)2', X).  X='uhlicitan vapenato-horecnaty'
%     dejnazev('UF6', X).         X='fluorid uranovy'
%
% Poznamky:
%     - Nazvoslovi beznych anorganickych sloucenin s urcitelnym oxidacnim cislem na urovni stredoskolske chemie
%     - Koordinacni slouceniny nejsou podporovany
%     - Krystalosolvaty (x-hydraty) nejsou podporovany
%     - Nazvoslovi samostatnych iontu neni podporovano
%     - Kovalentni hydridy nejsou prekladany podle normy
%     - Maximalni mozne cislo ve vzorci je 99
%
%
% Nektere prikladove slouceniny:
%     FeCl2, HCl, H2S, SO2, ZnO, O3, NaOH, Fe2O3, CO, CaMg(CO3)2
%     Na2O2, MgClOH, CaH2, N2H4, H3PO4, NaCl, FeCr2O4, Mg2TiO4, UF6
%     (SO3)3, SbCl5, AgN3, KCN, Ca(HCO3)2, BN, H2SO4, Ni2B, OsF8




dejnazev(ChemVyraz, Nazev):-
	xlate(ChemVyraz, Seznam),
	nazevsez(Seznam, NazevSeznam),
	name(Nazev, NazevSeznam).




% POMOCNE PREDIKATY

% conc( ?A, ?B, ?S ) - seznam S vznikne zretezenim A a B
conc(A, B, S):-append(A, B, S). % append je standartni predikat SWI-Prologu

% prvek( -Prvek, +Seznam)
% Priklad: prvek(h, [a,h,o,j]).
prvek(X, [X | _]).
prvek(X, [_ | T]):-prvek(X, T).

% ntyprvek(?N, ?Seznam, ?Prvek)
% Priklad: ntyprvek(c, 3, [a,b,c,d]).
ntyprvek(N, Seznam, Prvek):- nth1(N, Seznam, Prvek). % nth1 je standartni predikat SWI-Prologu

% uprav(+Seznam, +Krat, +Plus, -UpravenySeznam)
% Priklad: uprav( [4, 0, -6], 2, 1, [9, 1, -11]).
uprav([], _, _, []).
uprav( [Hlava1 | Ocas1], Krat, Plus, [Hlava2 | Ocas2] ):- 
	Hlava2 is Hlava1*Krat + Plus,
	uprav( Ocas1, Krat, Plus, Ocas2 ).





% PREDIKATY PRO PREKLAD CHEMICKYCH VYRAZU DO VNITRNIHO FORMATU

% xlate(+Vzorec, -Seznam)
% priklad: xlate('CaCO3', [ca, 1, c, 1, o, 3]).
xlate(Vzorec, VyslSeznam):-
	name(Vzorec, ASCIISez), 
	xlate2(ASCIISez, Seznam1), 
	xlate3(Seznam1, Seznam2),
	xlate4(Seznam2, VyslSeznam).

% xlate2(+ASCIISeznam, -Seznam) - prelozi seznam ASCII znaku puvodniho vyrazu na seznam pro dalsi upravu
% Priklad pro CaCO3: xlate2([67,97,67,79,51], [ca, c, o, 3]).
xlate2([], []).
xlate2([ASCIIVelke, ASCIIMale | ASCIIOcas], [Hlava | Ocas]):-
	ASCIIVelke >= 65, 
	ASCIIVelke =< 90,
	ASCIIMale >= 97,
	ASCIIMale =< 122,
	ASCIIZmensene is ASCIIVelke + 32,
	name(Hlava, [ASCIIZmensene, ASCIIMale]),
	xlate2( ASCIIOcas, Ocas ),!.
xlate2([ASCIIVelke | ASCIIOcas], [Hlava | Ocas]):-
	ASCIIVelke >= 65,
	ASCIIVelke =< 90,
	ASCIIZmensene is ASCIIVelke + 32,
	name(Hlava, [ASCIIZmensene]),
	xlate2(ASCIIOcas, Ocas).
xlate2([ASCIICislo | ASCIIOcas], [Hlava | Ocas]):-
	ASCIICislo >= 48,
	ASCIICislo =< 57,
	name(Hlava, [ASCIICislo]),
	xlate2(ASCIIOcas, Ocas).
xlate2([ASCIIPZavorka | ASCIIOcas], [PZavAtom | Ocas]):-
	ASCIIPZavorka =:= 40,
	name(PZavAtom, [ASCIIPZavorka]),
	xlate2(ASCIIOcas, Ocas).
xlate2([ASCIILZavorka | ASCIIOcas], [LZavAtom | Ocas]):-
	ASCIILZavorka =:= 41,
	name(LZavAtom, [ASCIILZavorka]),
	xlate2(ASCIIOcas, Ocas).

% xlate3(+VstupSez, -VystupSez) - nahradi zdvojena cisla jednim
% Priklad: xlate3([h,5,p,3,o,1,0], [h,5,p,3,o,10]).
xlate3([],[]).
xlate3([Cislo1, Cislo2 | Ocas1], [Cislo | Ocas2]):-
	integer(Cislo1),
	integer(Cislo2),
	Cislo is Cislo1*10 + Cislo2,
	xlate3(Ocas1, Ocas2).
xlate3([Neco | Ocas1], [Neco | Ocas2]):-
	xlate3(Ocas1, Ocas2).

% xlate4(+VstupSez, -VystupSez). - vlozi jednicky za prvky vyskytujici se ve vzorci jednou
% Priklad: xlate4([ca, c, o, 3], [ca, 1, c, 1, o, 3]).
xlate4([], []).
xlate4([Prvek, Cislo | Ocas1], [Prvek, Cislo | Ocas2]):- 
	atom(Prvek), 
	integer(Cislo),
	xlate4(Ocas1, Ocas2),
	!.
xlate4([PZavorka | Ocas1], [PZavorka | Ocas2]):-
	name(PZavorka, [40]),
	xlate4(Ocas1, Ocas2),
	!.
xlate4([Prvek1 | Ocas1], [Prvek1, 1 | Ocas2]):-
	atom(Prvek1),
	xlate4(Ocas1, Ocas2),
	!.
	


% DATABAZE

% Zjisteni kladnych oxid. cisel a nazvu prvku: data( ?Prvek, ?ZakladNazvu, -OxCisla)

data( li, [l,i,t],           [1] ):-!.          % lithium 
data( na, [s,o,d],           [1] ):-!.          % sodik
data( k,  [d,r,a,s,e,l],     [1] ):-!.          % draslik
data( rb, [r,u,b,i,d],       [1] ):-!.          % rubidium
data( cs, [c,e,s],           [1] ):-!.          % cesium
data( be, [b,e,r,y,l],       [2] ):-!.          % draslik
data( mg, [h,o,r,e,c],       [2] ):-!.          % horcik
data( ca, [v,a,p,e],         [2] ):-!.          % vapnik
data( sr, [s,t,r,o,n,e,c],   [2] ):-!.          % stroncium
data( ba, [b,a,r],           [2] ):-!.          % barium
data( u,  [u,r,a,n],         [2,3,4,5,6] ):-!.  % barium
data( ti, [t,i,t,a,n],       [3,4] ):-!.        % titan
data( cr, [c,h,r,o,m],       [2,3,6] ):-!.      % chrom
data( mo, [m,o,l,y,b,d,e,n], [4,6] ):-!.        % molybden
data( w,  [w,o,l,f,r,a,m],   [6] ):-!.          % wolfram
data( mn, [m,a,n,g,a,n],     [2,4,7] ):-!.      % mangan
data( fe, [z,e,l,e,z],       [2,3] ):-!.        % zelezo
data( os, [o,s,m],           [3,4,8] ):-!.      % osmium
data( co, [k,o,b,a,l],       [2,3] ):-!.        % kobalt
data( ni, [n,i,k,e,l],       [2,3] ):-!.        % nikl
data( pt, [p,l,a,t],         [2,4] ):-!.        % platina
data( cu, [m,e,d],           [1,2] ):-!.        % med
data( ag, [s,t,r,i,b,r],     [1] ):-!.          % stribro
data( au, [z,l,a,t],         [3] ):-!.          % zlato
data( zn, [z,i,n,e,c],       [2] ):-!.          % zinek
data( hg, [r,t,u,t],         [1,2] ):-!.        % rtut
data( b,  [b,o,r],           [3] ):-!.          % bor
data( al, [h,l,i,n],         [3] ):-!.          % hlinik
data( ga, [g,a,l],           [3] ):-!.          % galium
data( c,  [u,h,e,l],         [2,4] ):-!.        % uhlik
data( si, [k,r,e,m],         [4] ):-!.          % kremik
data( ge, [g,e,r,m,a,n],     [2,4] ):-!.        % germanium
data( sn, [c,i,n],           [2,4] ):-!.        % cin
data( pb, [o,l,o,v],         [2,4] ):-!.        % olovo
data( n,  [d,u,s],           [2,3,4,5] ):-!.    % dusik
data( p,  [f,o,s,f,o,r],     [1,3,5] ):-!.      % fosfor
data( as, [a,r,s,e,n],       [3,5] ):-!.        % arsen
data( sb, [a,n,t,i,m,o,n],   [3,5] ):-!.        % antimon
data( bi, [b,i,s,m,u,t],     [3] ):-!.          % bismut
data( s,  [s,i,r],           [4,6] ):-!.        % sira
data( se, [s,e,l,e,n],       [4,6] ):-!.        % selen
data( f,  [f,l,u,o,r],       [] ):-!.           % fluor
data( cl, [c,h,l,o,r],       [1,5,7] ):-!.      % chlor
data( br, [b,r,o,m],         [1,5] ):-!.        % brom
data( i,  [j,o,d],           [5,7] ):-!.        % jod


% zakonceni(?OxCislo, ?Seznam)
% Priklad: zakonceni(6, [o,v]).
zakonceni(0, []). % Specialni pripad
zakonceni(1, [n]).
zakonceni(2, [n,a,t]).
zakonceni(3, [i,t]).
zakonceni(4, [i,c,i,t]).
zakonceni(5, [e,c,n]).
zakonceni(6, [o,v]).
zakonceni(7, [i,s,t]).
zakonceni(8, [i,c,e,l]).


% predpona(?Cislo, ?Seznam)
% Priklad: predpona(3, [t,r,i]).
predpona(1, []).
predpona(2, [d,i]).
predpona(3, [t,r,i]).
predpona(4, [t,e,t,r,a]).
predpona(5, [p,e,n,t,a]).
predpona(6, [h,e,x,a]).
predpona(7, [h,e,p,t,a]).
predpona(8, [o,k,t,a]).
predpona(9, [n,o,n,a]).
predpona(10, [d,e,k,a]).
predpona(11, [u,n,d,e,k,a]).
predpona(12, [d,o,d,e,k,a]).
predpona(13, [t,r,i,d,e,k,a]).


% koncovka(+NazevAniontu, +Typ, -NazevAniontuEx)
% Priklad: koncovka([c,h,l,o,r], idskupina, [c,h,l,o,r,i,d]).
koncovka(Nazev, kysskupina, NazevEx) :- conc(Nazev, [a,n], NazevEx), !.
koncovka(Nazev, idskupina, NazevEx) :- conc(Nazev, [i,d], NazevEx), !.


% hydrogenpredpona(PocetH, PocetO, Seznam)
% Priklad pro H3PO4: hydrogenpredpona(3,4, [t,r,i,h,y,d,r,o,g,e,n]).
hydrogenpredpona(PocetH, PocetO, Seznam):-
	PocetH >= 3, PocetO >= 3,
	predpona(PocetH, Predp),
	conc(Predp, [h,y,d,r,o,g,e,n], Seznam),
	!.
hydrogenpredpona(_, _, []).



% nazvy(+ZakladNazvu, +SezOxCisel, -SeznamNazvu) - k ZakladNazvu prida pripony podle SezOxCisel
% Priklad: nazvy([s,i,r], [4,6], [[s,i,r,i,c,i,t],[s,i,r,o,v]]).
nazvy(_, [], []).
nazvy(ZakladNazvu, [OxCislo | Ocas1], [Nazev | Ocas2]):-	
	zakonceni(OxCislo, Zakon),
	conc(ZakladNazvu, Zakon, Nazev),
	nazvy(ZakladNazvu, Ocas1, Ocas2).



% znamaskupina(+Vzorec, -SeznamOxCisel, -SeznamNazvu, ?Typ)
% pro kazde mozne ox. cislo jeden nazev v seznamu seznamu
% Priklad: znamaskupina([s,1,o,4], [-4, -2], [[s,i,r,i,c,i,t],[s,i,r,o,v]], kysskupina).

znamaskupina([o,1,h,1], [-1],      [[h,y,d,r,o,x]],             idskupina).
znamaskupina([s,N],     [Minus2N], [[s,u,l,f]],                 idskupina):- Minus2N is 0- (2*N).
znamaskupina([h,1,s,1], [-1],      [[h,y,d,r,o,g,e,n,s,u,l,f]], idskupina). 
znamaskupina([c,1,n,1], [-1],      [[k,y,a,n]],                 idskupina).
znamaskupina([n,1,h,2], [-1],      [[a,m]],                     idskupina).
znamaskupina([n,3],     [-1],      [[a,z]],                     idskupina).
znamaskupina([n,N],     [Minus3N], [[n,i,t,r]],                 idskupina):- Minus3N is 0- (3*N).
znamaskupina([as,N],    [Minus3N], [[a,r,s,e,n]],               idskupina):- Minus3N is 0- (3*N).
znamaskupina([b,N],     [Minus3N], [[b,o,r]],                   idskupina):- Minus3N is 0- (3*N).
znamaskupina([c,N],     [MinusN],  [[k,a,r,b]],                 idskupina):- MinusN is 0 - N.
znamaskupina([h,N],     [MinusN],  [[h,y,d,r]],                 idskupina):- MinusN is 0 - N.
znamaskupina([o,2],     [-2],      [[p,e,r,o,x]],               idskupina).
znamaskupina([o,N],     [Minus2N], [[o,x]],                     idskupina):- Minus2N is 0 - 2*N.

% halogenidy
znamaskupina([i,X],     [MinusX],  [[j,o,d]],                   idskupina):- MinusX is 0 - X.
znamaskupina([f,X],     [MinusX],  [[f,l,u,o,r]],               idskupina):- MinusX is 0 - X.
znamaskupina([cl,X],    [MinusX],  [[c,h,l,o,r]],               idskupina):- MinusX is 0 - X.
znamaskupina([br,X],    [MinusX],  [[b,r,o,m]],                 idskupina):- MinusX is 0 - X.

% amonny kationt
znamaskupina([n,1,h,4], [1], [[a,m,o,n,n]], normal).


% kyselinotvorne skupiny
znamaskupina([M, Y, o, Z], SezOxCisel, SeznamNazvu, kysskupina):-
	integer(Y), integer(Z), atom(M),
	data(M, Zaklad, SezOxCiselM),
	Plus is 0- 2*Z,
	uprav(SezOxCiselM, Y, Plus, SezOxCisel),
	predpona(Y, Predpona),
	conc(Predpona, Zaklad, ZakladNazvu),
	nazvy(ZakladNazvu, SezOxCiselM, SeznamNazvu).	

znamaskupina( [h, X, M, Y, o, Z], SezOxCisel, SeznamNazvu, kysskupina):-
	integer(X), integer(Y), integer(Z),
	data(M, Zaklad, SezOxCiselM),
	predpona(X, Predpona1),
	predpona(Y, Predpona2),
	conc(Predpona1, [h,y,d,r,o,g,e,n], Sez1),
	conc(Predpona2, Zaklad, Sez2),
	conc(Sez1, Sez2, ZakladNazvu),	
	Plus is 0- 2*Z + X,
	uprav(SezOxCiselM, Y, Plus, SezOxCisel),	
	nazvy(ZakladNazvu, SezOxCiselM, SeznamNazvu).

% obecny prvek
znamaskupina([M, X], SezOxCisel, SezNazvu, normal):-
	integer(X),
	data(M, ZakladNazvu, OxCislaM),
	uprav(OxCislaM, X, 0, SezOxCisel),
	nazvy(ZakladNazvu, OxCislaM, SezNazvu).	

% skupina je v zavorce - zmenime jeji oxidacni cisla
znamaskupina(['(' | Ocas], OxCisla, SeznamNazvu, Typ):-
	conc(Podskupina, [')', X], Ocas),
	integer(X),
	znamaskupina(Podskupina, OxCislaPodskupiny, SeznamNazvu, Typ),
	uprav(OxCislaPodskupiny, X, 0, OxCisla).


	

% PREDIKAT PRO PREKLAD NAZVU

% nazevsez(+VzorecSeznam, -NazevSeznam)
% Priklad: nazevsez([h,2,s,1,o,4], [k,y,s,e,l,i,n,a,' ',s,i,r,o,v,a]).


% TRIVIALNI NAZVY NEKTERYCH SLOUCENIN
nazevsez([o,2], [k,y,s,l,i,k]):-!.
nazevsez([o,3], [o,z,o,n]):-!.
nazevsez([h,2,o,1], [v,o,d,a]):-!.
nazevsez([h,2,o,2], [p,e,r,o,x,i,d,' ',v,o,d,i,k,u]):-!.
nazevsez([n,1,h,3], [a,m,o,n,i,a,k]):-!.
nazevsez([n,2,h,4], [h,y,d,r,a,z,i,n]):-!.
nazevsez([h,1,n,3], [a,z,i,d,o,v,o,d,i,k]):-!.
nazevsez([h,1,f,1], [f,l,u,o,r,o,v,o,d,i,k]):-!.
nazevsez([h,1,cl,1], [c,h,l,o,r,o,v,o,d,i,k]):-!.
nazevsez([h,1,br,1], [b,r,o,m,o,v,o,d,i,k]):-!.
nazevsez([h,1,c,1,n,1], [k,y,a,n,o,v,o,d,i,k]):-!.
nazevsez([h,1,i,1], [j,o,d,o,v,o,d,i,k]):-!.
nazevsez([h,2,s,1], [s,i,r,o,v,o,d,i,k]):-!.


% Kyslikate kyseliny
% Typ HxMyOz, M je nejaky prvek
nazevsez([h,X,M,Y,o,Z], [k,y,s,e,l,i,n,a,' ' | Nazev]):-
	integer(X),
	znamaskupina([M, Y, o, Z], SeznamOxCisel, SeznamNazvu, kysskupina),
	MinusX is 0- X,
	ntyprvek(N, SeznamOxCisel, MinusX),
	ntyprvek(N, SeznamNazvu, ZakladNazvu),
	hydrogenpredpona(X, Z, Predpona),
	conc(Predpona, ZakladNazvu, Sez1),
	conc(Sez1, ['a'], Nazev).

% Prida x-merni priponu
nazevsez(['(' | Ocas], Nazev):-
	conc(Vnitrek, [')', X], Ocas),
	nazevsez(Vnitrek, NazevVnitrku),
	predpona(X, Predp),
	conc(NazevVnitrku, [' '], Sez1),
	conc(Predp, [m,e,r,n,i], Sez2),
	conc(Sez1, Sez2, Nazev).


% Slouceniny typu Kationt-Aniont
nazevsez(Vzorec, Nazev):-
	conc(Kationt, Aniont, Vzorec),
	znamaskupina(Kationt, SezOxCiselKationtu, SezNazvuKationtu, normal),
	znamaskupina(Aniont, SezOxCiselAniontu, SezNazvuAniontu, TypAniontu),
	ntyprvek(N1, SezOxCiselKationtu, OxCisloKationtu),
	OxCisloAniontu is 0 - OxCisloKationtu,
	ntyprvek(N2, SezOxCiselAniontu, OxCisloAniontu),
	ntyprvek(N1, SezNazvuKationtu, NazevKationtu),
	ntyprvek(N2, SezNazvuAniontu, NazevAniontu),
	koncovka(NazevAniontu, TypAniontu, NazevAniontuEx),
	conc(NazevAniontuEx, [' '], Sez1),
	conc(NazevKationtu, ['y'], Sez2),
	conc(Sez1, Sez2, Nazev).



% Slouceniny typu Kationt-Kationt-Aniont
nazevsez(Vzorec, Nazev):-
	conc(Kationty, Aniont, Vzorec), 
	conc(Kationt1, Kationt2, Kationty),
	znamaskupina(Kationt1, SezOxCiselKationtu1, SezNazvuKationtu1, normal),
	znamaskupina(Kationt2, SezOxCiselKationtu2, SezNazvuKationtu2, normal),
	znamaskupina(Aniont, SezOxCiselAniontu, SezNazvuAniontu, TypAniontu),
	ntyprvek(N1, SezOxCiselKationtu1, OxCisloKationtu1),
	ntyprvek(N2, SezOxCiselKationtu2, OxCisloKationtu2),
	OxCisloAniontu is 0 - OxCisloKationtu1 - OxCisloKationtu2,	
	ntyprvek(N3, SezOxCiselAniontu, OxCisloAniontu),	
	ntyprvek(N1, SezNazvuKationtu1, NazevKationtu1),
	ntyprvek(N2, SezNazvuKationtu2, NazevKationtu2),
	ntyprvek(N3, SezNazvuAniontu, NazevAniontu),
	koncovka(NazevAniontu, TypAniontu, NazevAniontuEx),
	conc(NazevAniontuEx, [' '], Sez1),
	conc(NazevKationtu1, [o,'-'], Sez2),
	conc(NazevKationtu2, [y], Sez3),
	conc(Sez1, Sez2, Sez12),
	conc(Sez12, Sez3, Nazev).



% Slouceniny typu Kationt-Aniont-Aniont
nazevsez(Vzorec, Nazev):-
	conc(Kationt, Anionty, Vzorec),
	conc(Aniont1, Aniont2, Anionty),
	znamaskupina(Kationt, SezOxCiselKationtu, SezNazvuKationtu, normal),
	znamaskupina(Aniont1, SezOxCiselAniontu1, SezNazvuAniontu1, TypAniontu1),
	znamaskupina(Aniont2, SezOxCiselAniontu2, SezNazvuAniontu2, TypAniontu2),	
	ntyprvek(N1, SezOxCiselAniontu1, OxCisloAniontu1),
	ntyprvek(N2, SezOxCiselAniontu2, OxCisloAniontu2),
	OxCisloKationtu is 0 - OxCisloAniontu1 - OxCisloAniontu2,
	ntyprvek(N3, SezOxCiselKationtu, OxCisloKationtu),
	ntyprvek(N1, SezNazvuAniontu1, NazevAniontu1),
	ntyprvek(N2, SezNazvuAniontu2, NazevAniontu2),
	ntyprvek(N3, SezNazvuKationtu, NazevKationtu),
	koncovka(NazevAniontu1, TypAniontu1, NazevAniontuEx1),
	koncovka(NazevAniontu2, TypAniontu2, NazevAniontuEx2),
	conc(NazevAniontuEx1, ['-'], Sez1),
	conc(NazevAniontuEx2, [' '], Sez2),
	conc(NazevKationtu, ['y'], Sez3),
	conc(Sez1, Sez2, Sez12),
	conc(Sez12, Sez3, Nazev).
















	
























